@@ -144,10 +144,11 @@ static GdkPixbuf *pixbuf_scale_to_fit  (GdkPixbuf  *src,
 static GdkPixbuf *pixbuf_scale_to_min  (GdkPixbuf  *src,
 					int         min_width,
 					int         min_height);
-static void       pixbuf_draw_gradient (GdkPixbuf  *pixbuf,
-					gboolean    horizontal,
-					GdkColor   *c1,
-					GdkColor   *c2);
+static void       pixbuf_draw_gradient (GdkPixbuf    *pixbuf,
+					gboolean      horizontal,
+					GdkColor     *c1,
+					GdkColor     *c2,
+					GdkRectangle *rect);
 static void       pixbuf_tile          (GdkPixbuf  *src,
 					GdkPixbuf  *dest);
 static void       pixbuf_blend         (GdkPixbuf  *src,
@@ -168,7 +169,9 @@ static gboolean   get_thumb_annotations (GdkPixbuf             *thumb,
 					 int                   *orig_height);
 
 /* Cache */
-static GdkPixbuf *get_pixbuf           (GnomeBG               *bg);
+static GdkPixbuf *get_pixbuf_for_size  (GnomeBG               *bg,
+					int                    width,
+					int                    height);
 static void       clear_cache          (GnomeBG               *bg);
 static gboolean   is_different         (GnomeBG               *bg,
 					const char            *filename);
@@ -183,7 +186,16 @@ static SlideShow * get_as_slideshow    (GnomeBG               *bg,
 					const char 	      *filename);
 static Slide *     get_current_slide   (SlideShow 	      *show,
 		   			double    	      *alpha);
-static gboolean    slideshow_changes_with_size (SlideShow *show);
+static gboolean    slideshow_has_multiple_sizes (SlideShow *show);
+
+static SlideShow *read_slideshow_file (const char *filename,
+				       GError     **err);
+static SlideShow *slideshow_ref       (SlideShow  *show);
+static void       slideshow_unref     (SlideShow  *show);
+
+static FileSize   *find_best_size      (GSList                *sizes,
+					gint                   width,
+					gint                   height);
 
 static void
 color_from_string (const char *string,
@@ -613,13 +625,15 @@ gnome_bg_set_filename (GnomeBG     *bg,
 }
 
 static void
-draw_color (GnomeBG *bg, GdkPixbuf *dest)
+draw_color_area (GnomeBG *bg,
+		 GdkPixbuf *dest,
+		 GdkRectangle *rect)
 {
 	guint32 pixel;
 	
-	switch (bg->color_type)
-	{
+	switch (bg->color_type) {
 	case GNOME_BG_COLOR_SOLID:
+		/* not really a big deal to ignore the area of interest */
 		pixel = ((bg->primary.red >> 8) << 24)      |
 			((bg->primary.green >> 8) << 16)    |
 			((bg->primary.blue >> 8) << 8)      |
@@ -629,11 +643,11 @@ draw_color (GnomeBG *bg, GdkPixbuf *dest)
 		break;
 		
 	case GNOME_BG_COLOR_H_GRADIENT:
-		pixbuf_draw_gradient (dest, TRUE, &(bg->primary), &(bg->secondary));
+		pixbuf_draw_gradient (dest, TRUE, &(bg->primary), &(bg->secondary), rect);
 		break;
 		
 	case GNOME_BG_COLOR_V_GRADIENT:
-		pixbuf_draw_gradient (dest, FALSE, &(bg->primary), &(bg->secondary));
+		pixbuf_draw_gradient (dest, FALSE, &(bg->primary), &(bg->secondary), rect);
 		break;
 		
 	default:
@@ -641,14 +655,77 @@ draw_color (GnomeBG *bg, GdkPixbuf *dest)
 	}
 }
 
+static void
+draw_color (GnomeBG *bg,
+	    GdkPixbuf *dest,
+	    GdkScreen *screen)
+{
+	GdkRectangle rect;
+	rect.x = 0;
+	rect.y = 0;
+	rect.width = gdk_pixbuf_get_width (dest);
+	rect.height = gdk_pixbuf_get_height (dest);
+	draw_color_area (bg, dest, &rect);
+}
+
+static void
+draw_color_each_monitor (GnomeBG *bg,
+			 GdkPixbuf *dest,
+			 GdkScreen *screen)
+{
+	GdkRectangle rect;
+	gint num_monitors;
+	int monitor;
+
+	num_monitors = gdk_screen_get_n_monitors (screen);
+	for (monitor = 0; monitor < num_monitors; monitor++) {
+		gdk_screen_get_monitor_geometry (screen, monitor, &rect);
+		draw_color_area (bg, dest, &rect);
+	}
+}
+
+static GdkPixbuf *
+pixbuf_clip_to_fit (GdkPixbuf *src,
+		    int        max_width,
+		    int        max_height)
+{
+	int src_width, src_height;
+	int w, h;
+	int src_x, src_y;
+	GdkPixbuf *pixbuf;
+
+	src_width = gdk_pixbuf_get_width (src);
+	src_height = gdk_pixbuf_get_height (src);
+
+	if (src_width < max_width && src_height < max_height)
+		return g_object_ref (src);
+
+	w = MIN(src_width, max_width);
+	h = MIN(src_height, max_height);
+
+	pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
+				 gdk_pixbuf_get_has_alpha (src),
+				 8, w, h);
+
+	src_x = (src_width - w) / 2;
+	src_y = (src_height - h) / 2;
+	gdk_pixbuf_copy_area (src,
+			      src_x, src_y,
+			      w, h,
+			      pixbuf,
+			      0, 0);
+	return pixbuf;
+}
+
 static GdkPixbuf *
 get_scaled_pixbuf (GnomeBGPlacement placement,
 		   GdkPixbuf *pixbuf,
 		   int width, int height,
-		   int *x, int *y, int *w, int *h)
+		   int *x, int *y,
+		   int *w, int *h)
 {
 	GdkPixbuf *new;
-	
+
 #if 0
 	g_print ("original_width: %d %d\n",
 		 gdk_pixbuf_get_width (pixbuf),
@@ -672,7 +749,7 @@ get_scaled_pixbuf (GnomeBGPlacement placement,
 	case GNOME_BG_PLACEMENT_CENTERED:
 	case GNOME_BG_PLACEMENT_TILED:
 	default:
-		new = g_object_ref (pixbuf);
+		new = pixbuf_clip_to_fit (pixbuf, width, height);
 		break;
 	}
 	
@@ -685,20 +762,20 @@ get_scaled_pixbuf (GnomeBGPlacement placement,
 }
 
 static void
-draw_image (GnomeBGPlacement  placement,
-	    GdkPixbuf        *pixbuf,
-	    GdkPixbuf        *dest)
+draw_image_area (GnomeBGPlacement  placement,
+		 GdkPixbuf        *pixbuf,
+		 GdkPixbuf        *dest,
+		 GdkRectangle     *area)
 {
-	int dest_width = gdk_pixbuf_get_width (dest);
-	int dest_height = gdk_pixbuf_get_height (dest);
+	int dest_width = area->width;
+	int dest_height = area->height;
 	int x, y, w, h;
 	GdkPixbuf *scaled;
 	
 	if (!pixbuf)
 		return;
-	
-	scaled = get_scaled_pixbuf (
-		placement, pixbuf, dest_width, dest_height, &x, &y, &w, &h);
+
+	scaled = get_scaled_pixbuf (placement, pixbuf, dest_width, dest_height, &x, &y, &w, &h);
 
 	switch (placement) {
 	case GNOME_BG_PLACEMENT_TILED:
@@ -708,7 +785,7 @@ draw_image (GnomeBGPlacement  placement,
 	case GNOME_BG_PLACEMENT_CENTERED:
 	case GNOME_BG_PLACEMENT_FILL_SCREEN:
 	case GNOME_BG_PLACEMENT_SCALED:
-		pixbuf_blend (scaled, dest, 0, 0, w, h, x, y, 1.0);
+		pixbuf_blend (scaled, dest, 0, 0, w, h, x + area->x, y + area->y, 1.0);
 		break;
 	default:
 		g_assert_not_reached ();
@@ -718,54 +795,112 @@ draw_image (GnomeBGPlacement  placement,
 	g_object_unref (scaled);
 }
 
-void
-gnome_bg_draw (GnomeBG *bg, GdkPixbuf *dest)
+static void
+draw_image (GnomeBGPlacement  placement,
+	    GdkPixbuf        *pixbuf,
+	    GdkPixbuf        *dest)
 {
-	if (!bg)
-		return;
-	
-	draw_color (bg, dest);
-	
-	draw_image (bg->placement, get_pixbuf (bg), dest);
+	GdkRectangle rect;
+
+	rect.x = 0;
+	rect.y = 0;
+	rect.width = gdk_pixbuf_get_width (dest);
+	rect.height = gdk_pixbuf_get_height (dest);
+
+	draw_image_area (placement, pixbuf, dest, &rect);
 }
 
-gboolean
-gnome_bg_changes_with_size (GnomeBG *bg)
+static void
+draw_once (GnomeBG   *bg,
+	   GdkPixbuf *dest,
+	   GdkScreen *screen)
 {
-	SlideShow *show;
+	GdkRectangle rect;
+	GdkPixbuf   *pixbuf;
 
-	g_return_val_if_fail (bg != NULL, FALSE);
+	rect.x = 0;
+	rect.y = 0;
+	rect.width = gdk_pixbuf_get_width (dest);
+	rect.height = gdk_pixbuf_get_height (dest);
 
-	show = get_as_slideshow (bg, bg->filename);
-	if (show) 
-		return slideshow_changes_with_size (show);
-	
-	if (bg->color_type != GNOME_BG_COLOR_SOLID) {
-		if (!get_pixbuf (bg))
-			return TRUE;
-		if (gdk_pixbuf_get_has_alpha (get_pixbuf (bg)))
-			return TRUE;
-		if (bg->placement == GNOME_BG_PLACEMENT_CENTERED)
-			return TRUE;
-		return FALSE;
-	}
-	else if (bg->placement == GNOME_BG_PLACEMENT_TILED) {
-		return FALSE;
-	}
-	else {
-		return TRUE;
+	pixbuf = get_pixbuf_for_size (bg, gdk_pixbuf_get_width (dest), gdk_pixbuf_get_height (dest));
+	if (pixbuf) {
+		draw_image_area (bg->placement,
+				 pixbuf,
+				 dest,
+				 &rect);
+		g_object_unref (pixbuf);
 	}
 }
 
 static void
-gnome_bg_get_pixmap_size (GnomeBG  *bg,
-			  int       width,
-			  int       height,
-			  int      *pixmap_width,
-			  int      *pixmap_height)
+draw_each_monitor (GnomeBG   *bg,
+		   GdkPixbuf *dest,
+		   GdkScreen *screen)
+{
+	GdkRectangle rect;
+	gint num_monitors;
+	int monitor;
+
+	num_monitors = gdk_screen_get_n_monitors (screen);
+	for (monitor = 0; monitor < num_monitors; monitor++) {
+		GdkPixbuf *pixbuf;
+		gdk_screen_get_monitor_geometry (screen, monitor, &rect);
+		pixbuf = get_pixbuf_for_size (bg, rect.width, rect.height);
+		if (pixbuf) {
+			draw_image_area (bg->placement,
+					 pixbuf,
+					 dest, &rect);
+			g_object_unref (pixbuf);
+		}
+	}
+}
+
+void
+gnome_bg_draw (GnomeBG *bg,
+	       GdkPixbuf *dest,
+	       GdkScreen *screen,
+	       gboolean is_root)
+{
+	if (!bg)
+		return;
+
+	if (is_root) {
+		draw_color_each_monitor (bg, dest, screen);
+		draw_each_monitor (bg, dest, screen);
+	} else {
+		draw_color (bg, dest, screen);
+		draw_once (bg, dest, screen);
+	}
+}
+
+gboolean
+gnome_bg_has_multiple_sizes (GnomeBG *bg)
+{
+	SlideShow *show;
+	gboolean ret;
+
+	g_return_val_if_fail (bg != NULL, FALSE);
+
+	ret = FALSE;
+
+	show = get_as_slideshow (bg, bg->filename);
+	if (show) {
+		ret = slideshow_has_multiple_sizes (show);
+		slideshow_unref (show);
+	}
+
+	return ret;
+}
+
+static void
+gnome_bg_get_pixmap_size (GnomeBG   *bg,
+			  int        width,
+			  int        height,
+			  int       *pixmap_width,
+			  int       *pixmap_height)
 {
 	int dummy;
-	int pb_width, pb_height;
 	
 	if (!pixmap_width)
 		pixmap_width = &dummy;
@@ -774,8 +909,8 @@ gnome_bg_get_pixmap_size (GnomeBG  *bg,
 	
 	*pixmap_width = width;
 	*pixmap_height = height;
-	
-	if (!get_pixbuf (bg)) {
+
+	if (!bg->filename) {
 		switch (bg->color_type) {
 		case GNOME_BG_COLOR_SOLID:
 			*pixmap_width = 1;
@@ -783,44 +918,12 @@ gnome_bg_get_pixmap_size (GnomeBG  *bg,
 			break;
 			
 		case GNOME_BG_COLOR_H_GRADIENT:
-			*pixmap_width = width;
-			*pixmap_height = GRADIENT_PIXMAP_TILE_SIZE;
-			break;
-			
 		case GNOME_BG_COLOR_V_GRADIENT:
-			*pixmap_width = GRADIENT_PIXMAP_TILE_SIZE;
-			*pixmap_height = height;
 			break;
 		}
 		
 		return;
 	}
-	
-	pb_width = gdk_pixbuf_get_width (get_pixbuf (bg));
-	pb_height = gdk_pixbuf_get_height (get_pixbuf (bg));
-	
-	if (bg->placement == GNOME_BG_PLACEMENT_TILED) {
-		if (gdk_pixbuf_get_has_alpha (get_pixbuf (bg)) &&
-		    bg->color_type != GNOME_BG_COLOR_SOLID) {
-			if (bg->color_type == GNOME_BG_COLOR_H_GRADIENT) {
-				/* FIXME: Should this be
-				 * MAX (GRADIENT_TILE_SIZE, pb_height)?
-				 */
-				*pixmap_height = pb_height; 
-				*pixmap_width = width;
-			}
-			else {
-				/* FIXME: Should this be
-				 * MAX (GRAIDENT_TILE_SIZE, pb_width? */
-				*pixmap_width = pb_width;
-				*pixmap_height = height;
-			}
-		}
-		else {
-			*pixmap_width = pb_width;
-			*pixmap_height = pb_height;
-		}
-	}
 }
 
 /**
@@ -842,7 +945,7 @@ gnome_bg_create_pixmap (GnomeBG	    *bg,
 			GdkWindow   *window,
 			int	     width,
 			int	     height,
-			gboolean     root)
+			gboolean     is_root)
 {
 	int pm_width, pm_height;
 	GdkPixmap *pixmap;
@@ -860,9 +963,10 @@ gnome_bg_create_pixmap (GnomeBG	    *bg,
 	bg->last_pixmap_width = width;
 	bg->last_pixmap_height = height;
 
+	/* has the side effect of loading and caching pixbuf only when in tile mode */
 	gnome_bg_get_pixmap_size (bg, width, height, &pm_width, &pm_height);
 	
-	if (root) {
+	if (is_root) {
 		pixmap = make_root_pixmap (gdk_drawable_get_screen (window),
 					   pm_width, pm_height);
 	}
@@ -870,7 +974,7 @@ gnome_bg_create_pixmap (GnomeBG	    *bg,
 		pixmap = gdk_pixmap_new (window, pm_width, pm_height, -1);
 	}
 	
-	if (!get_pixbuf (bg) && bg->color_type == GNOME_BG_COLOR_SOLID) {
+	if (!bg->filename && bg->color_type == GNOME_BG_COLOR_SOLID) {
 		GdkGC *gc = gdk_gc_new (pixmap);
 		gdk_gc_set_rgb_fg_color (gc, &(bg->primary));
 		
@@ -883,7 +987,7 @@ gnome_bg_create_pixmap (GnomeBG	    *bg,
 		
 		pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8,
 					 width, height);
-		gnome_bg_draw (bg, pixbuf);
+		gnome_bg_draw (bg, pixbuf, gdk_drawable_get_screen (GDK_DRAWABLE (window)), is_root);
 		gdk_draw_pixbuf (pixmap, NULL, pixbuf,
 				 0, 0,
 				 0, 0, width, height,
@@ -899,10 +1003,13 @@ gnome_bg_create_pixmap (GnomeBG	    *bg,
  * clients know what colors to draw on top with
  */
 gboolean
-gnome_bg_is_dark (GnomeBG *bg)
+gnome_bg_is_dark (GnomeBG *bg,
+		  int      width,
+		  int      height)
 {
 	GdkColor color;
 	int intensity;
+	GdkPixbuf *pixbuf;
 	
 	g_return_val_if_fail (bg != NULL, FALSE);
 	
@@ -913,9 +1020,9 @@ gnome_bg_is_dark (GnomeBG *bg)
 		color.green = (bg->primary.green + bg->secondary.green) / 2;
 		color.blue = (bg->primary.blue + bg->secondary.blue) / 2;
 	}
-	
-	if (get_pixbuf (bg)) {
-		guint32 argb = pixbuf_average_value (get_pixbuf (bg));
+	pixbuf = get_pixbuf_for_size (bg, width, height);
+	if (pixbuf) {
+		guint32 argb = pixbuf_average_value (pixbuf);
 		guchar a = (argb >> 24) & 0xff;
 		guchar r = (argb >> 16) & 0xff;
 		guchar g = (argb >>  8) & 0xff;
@@ -924,6 +1031,7 @@ gnome_bg_is_dark (GnomeBG *bg)
 		color.red = (color.red * (0xFF - a) + r * 0x101 * a) / 0xFF;
 		color.green = (color.green * (0xFF - a) + g * 0x101 * a) / 0xFF;
 		color.blue = (color.blue * (0xFF - a) + b * 0x101 * a) / 0xFF;
+		g_object_unref (pixbuf);
 	}
 	
 	intensity = (color.red * 77 +
@@ -1001,9 +1109,32 @@ get_original_size (const char *filename,
 	return result;
 }
 
+static const char *
+get_filename_for_size (GnomeBG *bg, gint best_width, gint best_height)
+{
+	SlideShow *show;
+	Slide *slide;
+	FileSize *size;
+
+	if (!bg->filename)
+		return NULL;
+
+	show = get_as_slideshow (bg, bg->filename);
+	if (!show) {
+		return bg->filename;
+	}
+
+	slide = get_current_slide (show, NULL);
+	slideshow_unref (show);
+	size = find_best_size (slide->file1, best_width, best_height);
+	return size->file;
+}
+
 gboolean
 gnome_bg_get_image_size (GnomeBG	       *bg,
 			 GnomeDesktopThumbnailFactory *factory,
+			 int                    best_width,
+			 int                    best_height,
 			 int		       *width,
 			 int		       *height)
 {
@@ -1017,21 +1148,8 @@ gnome_bg_get_image_size (GnomeBG	       *bg,
 	if (!bg->filename)
 		return FALSE;
 	
-	filename = bg->filename;
+	filename = get_filename_for_size (bg, best_width, best_height);
 	thumb = create_thumbnail_for_filename (factory, filename);
-	
-	if (!thumb) {
-		SlideShow *show = get_as_slideshow (bg, bg->filename);
-		if (show) {
-			double alpha;
-			FileSize *fs;
-			Slide *slide = get_current_slide (show, &alpha);
-			fs = slide->file1->data;
-			filename = fs->file;
-			thumb = create_thumbnail_for_filename (factory, filename);
-		}
-	}
-
 	if (thumb) {
 		if (get_thumb_annotations (thumb, width, height))
 			result = TRUE;
@@ -1068,7 +1186,7 @@ gnome_bg_create_thumbnail (GnomeBG               *bg,
 	
 	result = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, dest_width, dest_height);
 	
-	draw_color (bg, result);
+	draw_color (bg, result, screen);
 	
 	thumb = create_img_thumbnail (bg, factory, screen, dest_width, dest_height, -1);
 	
@@ -1316,17 +1434,13 @@ struct _SlideShow
 
 	GQueue *slides;
 	
-	gboolean changes_with_size;
+	gboolean has_multiple_sizes;
 
 	/* used during parsing */
 	struct tm start_tm;
 	GQueue *stack;
 };
 
-static SlideShow *read_slideshow_file (const char *filename,
-				       GError     **err);
-static void       slideshow_ref       (SlideShow  *show);
-static void       slideshow_unref     (SlideShow  *show);
 
 static double
 now (void)
@@ -1356,7 +1470,8 @@ get_current_slide (SlideShow *show,
 		Slide *slide = list->data;
 
 		if (elapsed + slide->duration > delta) {
-			*alpha = (delta - elapsed) / (double)slide->duration;
+			if (alpha)
+				*alpha = (delta - elapsed) / (double)slide->duration;
 			return slide;
 		}
 
@@ -1441,9 +1556,9 @@ bound_cache (GnomeBG *bg)
       while (g_list_length (bg->file_cache) >= CACHE_SIZE) {
 	      GList *last_link = g_list_last (bg->file_cache);
 	      FileCacheEntry *ent = last_link->data;
-	      
+
 	      file_cache_entry_delete (ent);
-	      
+
 	      bg->file_cache = g_list_delete_link (bg->file_cache, last_link);
       }
 }
@@ -1490,7 +1605,7 @@ file_cache_add_pixbuf (GnomeBG *bg,
 		       GdkPixbuf *pixbuf)
 {
 	FileCacheEntry *ent = file_cache_entry_new (bg, PIXBUF, filename);
-	ent->u.pixbuf = pixbuf;
+	ent->u.pixbuf = g_object_ref (pixbuf);
 }
 
 static void
@@ -1499,7 +1614,7 @@ file_cache_add_thumbnail (GnomeBG *bg,
 			  GdkPixbuf *pixbuf)
 {
 	FileCacheEntry *ent = file_cache_entry_new (bg, THUMBNAIL, filename);
-	ent->u.thumbnail = pixbuf;
+	ent->u.thumbnail = g_object_ref (pixbuf);
 }
 
 static void
@@ -1508,29 +1623,32 @@ file_cache_add_slide_show (GnomeBG *bg,
 			   SlideShow *show)
 {
 	FileCacheEntry *ent = file_cache_entry_new (bg, SLIDESHOW, filename);
-	ent->u.slideshow = show;
+	ent->u.slideshow = slideshow_ref (show);
 }
 
 static GdkPixbuf *
-get_as_pixbuf (GnomeBG *bg, const char *filename)
+get_as_pixbuf_for_size (GnomeBG    *bg,
+			const char *filename,
+			int         best_width,
+			int         best_height)
 {
 	const FileCacheEntry *ent;
 	if ((ent = file_cache_lookup (bg, PIXBUF, filename))) {
-		return ent->u.pixbuf;
+		return g_object_ref (ent->u.pixbuf);
 	}
 	else {
 		GdkPixbufFormat *format;
 		GdkPixbuf *pixbuf;
 
 		/* If scalable choose maximum size */
-		format = gdk_pixbuf_get_file_info (bg->filename, NULL, NULL);
+		format = gdk_pixbuf_get_file_info (filename, NULL, NULL);
 		if (format != NULL &&
 		    strcmp (gdk_pixbuf_format_get_name (format), "svg") == 0 &&
-		    (bg->last_pixmap_width > 0 && bg->last_pixmap_height > 0) &&
+		    (best_width > 0 && best_height > 0) &&
 		    (bg->placement == GNOME_BG_PLACEMENT_FILL_SCREEN ||
 		     bg->placement == GNOME_BG_PLACEMENT_SCALED ||
 		     bg->placement == GNOME_BG_PLACEMENT_ZOOMED))
-			pixbuf = gdk_pixbuf_new_from_file_at_size (filename, bg->last_pixmap_width, bg->last_pixmap_height, NULL);
+			pixbuf = gdk_pixbuf_new_from_file_at_size (filename, best_width, best_height, NULL);
 		else
 			pixbuf = gdk_pixbuf_new_from_file (filename, NULL);
 
@@ -1546,7 +1664,7 @@ get_as_slideshow (GnomeBG *bg, const char *filename)
 {
 	const FileCacheEntry *ent;
 	if ((ent = file_cache_lookup (bg, SLIDESHOW, filename))) {
-		return ent->u.slideshow;
+		return slideshow_ref (ent->u.slideshow);
 	}
 	else {
 		SlideShow *show = read_slideshow_file (filename, NULL);
@@ -1563,7 +1681,7 @@ get_as_thumbnail (GnomeBG *bg, GnomeDesktopThumbnailFactory *factory, const char
 {
 	const FileCacheEntry *ent;
 	if ((ent = file_cache_lookup (bg, THUMBNAIL, filename))) {
-		return ent->u.thumbnail;
+		return g_object_ref (ent->u.thumbnail);
 	}
 	else {
 		GdkPixbuf *thumb = create_thumbnail_for_filename (factory, filename);
@@ -1760,22 +1878,28 @@ create_img_thumbnail (GnomeBG                      *bg,
 		      int                           frame_num)
 {
 	if (bg->filename) {
-		GdkPixbuf *thumb = get_as_thumbnail (bg, factory, bg->filename);
+		GdkPixbuf *thumb;
+
+		thumb = get_as_thumbnail (bg, factory, bg->filename);
 
 		if (thumb) {
-			return scale_thumbnail (
-				bg->placement, bg->filename,
-				thumb, screen, dest_width, dest_height);
+			GdkPixbuf *result;
+			result = scale_thumbnail (bg->placement,
+						  bg->filename,
+						  thumb,
+						  screen,
+						  dest_width,
+						  dest_height);
+			g_object_unref (thumb);
+			return result;
 		}
 		else {
 			SlideShow *show = get_as_slideshow (bg, bg->filename);
-			
+
 			if (show) {
 				double alpha;
 				Slide *slide;
 
-				slideshow_ref (show);
-
 				if (frame_num == -1)
 					slide = get_current_slide (show, &alpha);
 				else
@@ -1784,42 +1908,53 @@ create_img_thumbnail (GnomeBG                      *bg,
 				if (slide->fixed) {
 					GdkPixbuf *tmp;
 					FileSize *fs;
-					
-					fs = slide->file1->data;
+					fs = find_best_size (slide->file1, dest_width, dest_height);
 					tmp = get_as_thumbnail (bg, factory, fs->file);
-
-					thumb = scale_thumbnail (
-						bg->placement, fs->file,
-						tmp, screen, dest_width, dest_height);
+					if (tmp) {
+						thumb = scale_thumbnail (bg->placement,
+									 fs->file,
+									 tmp,
+									 screen,
+									 dest_width,
+									 dest_height);
+						g_object_unref (tmp);
+					}
 				}
 				else {
-					FileSize *fs;
+					FileSize *fs1, *fs2;
 					GdkPixbuf *p1, *p2;
+					fs1 = find_best_size (slide->file1, dest_width, dest_height);
+					p1 = get_as_thumbnail (bg, factory, fs1->file);
 
-					fs = slide->file1->data;
-					p1 = get_as_thumbnail (bg, factory, fs->file);
-
-					fs = slide->file2->data;
-					p2 = get_as_thumbnail (bg, factory, fs->file);
+					fs2 = find_best_size (slide->file2, dest_width, dest_height);
+					p2 = get_as_thumbnail (bg, factory, fs2->file);
 
 					if (p1 && p2) {
 						GdkPixbuf *thumb1, *thumb2;
 
-						fs = slide->file1->data;
-						thumb1 = scale_thumbnail (
-							bg->placement, fs->file,
-							p1, screen, dest_width, dest_height);
+						thumb1 = scale_thumbnail (bg->placement,
+									  fs1->file,
+									  p1,
+									  screen,
+									  dest_width,
+									  dest_height);
 
-						fs = slide->file2->data;
-						thumb2 = scale_thumbnail (
-							bg->placement, fs->file,
-							p2, screen, dest_width, dest_height);
+						thumb2 = scale_thumbnail (bg->placement,
+									  fs2->file,
+									  p2,
+									  screen,
+									  dest_width,
+									  dest_height);
 
 						thumb = blend (thumb1, thumb2, alpha);
 
 						g_object_unref (thumb1);
 						g_object_unref (thumb2);
 					}
+					if (p1)
+						g_object_unref (p1);
+					if (p2)
+						g_object_unref (p2);
 				}
 
 				ensure_timeout (bg, slide);
@@ -1880,18 +2015,29 @@ find_best_size (GSList *sizes, gint width, gint height)
 }
 
 static GdkPixbuf *
-get_pixbuf (GnomeBG *bg)
+get_pixbuf_for_size (GnomeBG *bg,
+		     gint best_width,
+		     gint best_height)
 {
-	/* FIXME: this ref=TRUE/FALSE stuff is crazy */
-	
 	guint time_until_next_change;
-	gboolean ref = FALSE;
-	
-	if (!bg->pixbuf_cache && bg->filename) {
-		ref = TRUE;
+	gboolean hit_cache = FALSE;
+
+	/* only hit the cache if the aspect ratio matches */
+	if (bg->pixbuf_cache) {
+		int width, height;
+		width = gdk_pixbuf_get_width (bg->pixbuf_cache);
+		height = gdk_pixbuf_get_height (bg->pixbuf_cache);
+		hit_cache = 0.2 > fabs ((best_width / (double)best_height) - (width / (double)height));
+		if (!hit_cache) {
+			g_object_unref (bg->pixbuf_cache);
+			bg->pixbuf_cache = NULL;
+		}
+	}
+
+	if (!hit_cache && bg->filename) {
 		bg->file_mtime = get_mtime (bg->filename);
-		
-		bg->pixbuf_cache = get_as_pixbuf (bg, bg->filename);
+
+		bg->pixbuf_cache = get_as_pixbuf_for_size (bg, bg->filename, best_width, best_height);
 		time_until_next_change = G_MAXUINT;
 		if (!bg->pixbuf_cache) {
 			SlideShow *show = get_as_slideshow (bg, bg->filename);
@@ -1906,22 +2052,24 @@ get_pixbuf (GnomeBG *bg)
 				time_until_next_change = (guint)get_slide_timeout (slide);
 				if (slide->fixed) {
 					FileSize *size;
-					size = find_best_size (slide->file1, bg->last_pixmap_width, bg->last_pixmap_height);
-					bg->pixbuf_cache = get_as_pixbuf (bg, size->file);
+					size = find_best_size (slide->file1, best_width, best_height);
+					bg->pixbuf_cache = get_as_pixbuf_for_size (bg, size->file, best_width, best_height);
 				}
 				else {
 					FileSize *size;
 					GdkPixbuf *p1, *p2;
-					size = find_best_size (slide->file1, bg->last_pixmap_width, bg->last_pixmap_height);
-					p1 = get_as_pixbuf (bg, size->file);
-					size = find_best_size (slide->file2, bg->last_pixmap_width, bg->last_pixmap_height);
-					p2 = get_as_pixbuf (bg, size->file);
-
+					size = find_best_size (slide->file1, best_width, best_height);
+					p1 = get_as_pixbuf_for_size (bg, size->file, best_width, best_height);
+					size = find_best_size (slide->file2, best_width, best_height);
+					p2 = get_as_pixbuf_for_size (bg, size->file, best_width, best_height);
 
 					if (p1 && p2) {
 						bg->pixbuf_cache = blend (p1, p2, alpha);
-						ref = FALSE;
 					}
+					if (p1)
+						g_object_unref (p1);
+					if (p2)
+						g_object_unref (p2);
 				}
 
 				ensure_timeout (bg, slide);
@@ -1937,9 +2085,9 @@ get_pixbuf (GnomeBG *bg)
 		    blow_expensive_caches_in_idle (bg);
 	}
 
-	if (bg->pixbuf_cache && ref)
+	if (bg->pixbuf_cache)
 		g_object_ref (bg->pixbuf_cache);
-	
+
 	return bg->pixbuf_cache;
 }
 
@@ -2084,22 +2232,38 @@ pixbuf_scale_to_min (GdkPixbuf *src, int min_width, int min_height)
 	double factor;
 	int src_width, src_height;
 	int new_width, new_height;
-	
+	GdkPixbuf *dest;
+
 	src_width = gdk_pixbuf_get_width (src);
 	src_height = gdk_pixbuf_get_height (src);
-	
+
 	factor = MAX (min_width / (double) src_width, min_height / (double) src_height);
-	
+
 	new_width = floor (src_width * factor + 0.5);
 	new_height = floor (src_height * factor + 0.5);
-	
-	return gdk_pixbuf_scale_simple (src, new_width, new_height, GDK_INTERP_BILINEAR);
+
+	dest = gdk_pixbuf_new (GDK_COLORSPACE_RGB,
+			       gdk_pixbuf_get_has_alpha (src),
+			       8, min_width, min_height);
+	if (!dest)
+		return NULL;
+
+	/* crop the result */
+	gdk_pixbuf_scale (src, dest,
+			  0, 0,
+			  min_width, min_height,
+			  (new_width - min_width) / -2,
+			  (new_height - min_height) / -2,
+			  factor,
+			  factor,
+			  GDK_INTERP_BILINEAR);
+	return dest;
 }
 
 static guchar *
 create_gradient (const GdkColor *primary,
 		 const GdkColor *secondary,
-		 int	          n_pixels)
+		 int	         n_pixels)
 {
 	guchar *result = g_malloc (n_pixels * 3);
 	int i;
@@ -2116,43 +2280,56 @@ create_gradient (const GdkColor *primary,
 }	
 
 static void
-pixbuf_draw_gradient (GdkPixbuf *pixbuf,
-		      gboolean   horizontal,
-		      GdkColor  *primary,
-		      GdkColor  *secondary)
+pixbuf_draw_gradient (GdkPixbuf    *pixbuf,
+		      gboolean      horizontal,
+		      GdkColor     *primary,
+		      GdkColor     *secondary,
+		      GdkRectangle *rect)
 {
-	int width  = gdk_pixbuf_get_width (pixbuf);
-	int height = gdk_pixbuf_get_height (pixbuf);
-	int rowstride = gdk_pixbuf_get_rowstride (pixbuf);
-	guchar *dst = gdk_pixbuf_get_pixels (pixbuf);
-	guchar *dst_limit = dst + height * rowstride;
-	
+	int width;
+	int height;
+	int rowstride;
+	guchar *dst;
+	guchar *dst_limit;
+	int n_channels = 3;
+
+	rowstride = gdk_pixbuf_get_rowstride (pixbuf);
+	width = rect->width;
+	height = rect->height;
+	dst = gdk_pixbuf_get_pixels (pixbuf) + rect->x * n_channels + rowstride * rect->y;
+	dst_limit = dst + height * rowstride;
+
 	if (horizontal) {
 		guchar *gradient = create_gradient (primary, secondary, width);
-		int copy_bytes_per_row = width * 3;
-		
-		while (dst < dst_limit) {
-			memcpy (dst, gradient, copy_bytes_per_row);
-			dst += rowstride;
+		int copy_bytes_per_row = width * n_channels;
+		int i;
+
+		for (i = 0; i < height; i++) {
+			guchar *d;
+			d = dst + rowstride * i;
+			memcpy (d, gradient, copy_bytes_per_row);
 		}
 		g_free (gradient);
 	} else {
 		guchar *gb, *gradient;
-		
-		gb = gradient = create_gradient (primary, secondary, height);
-		while (dst < dst_limit) {
-			int i;
-			guchar *d = dst;
-			guchar r = *gb++;
-			guchar g = *gb++;
-			guchar b = *gb++;
-			for (i = 0; i < width; i++) {
-				*d++ = r;
-				*d++ = g;
-				*d++ = b;
+		int i;
+
+		gradient = create_gradient (primary, secondary, height);
+		for (i = 0; i < height; i++) {
+			int j;
+			guchar *d;
+
+			d = dst + rowstride * i;
+			gb = gradient + n_channels * i;
+			for (j = width; j > 0; j--) {
+				int k;
+
+				for (k = 0; k < n_channels; k++) {
+					*(d++) = gb[k];
+				}
 			}
-			dst += rowstride;
 		}
+
 		g_free (gradient);
 	}
 }
@@ -2162,8 +2339,8 @@ pixbuf_blend (GdkPixbuf *src,
 	      GdkPixbuf *dest,
 	      int	 src_x,
 	      int	 src_y,
-	      int	 width,
-	      int        height,
+	      int	 src_width,
+	      int        src_height,
 	      int	 dest_x,
 	      int	 dest_y,
 	      double	 alpha)
@@ -2173,11 +2350,11 @@ pixbuf_blend (GdkPixbuf *src,
 	int offset_x = dest_x - src_x;
 	int offset_y = dest_y - src_y;
 
-	if (width < 0)
-		width = gdk_pixbuf_get_width (src);
+	if (src_width < 0)
+		src_width = gdk_pixbuf_get_width (src);
 
-	if (height < 0)
-		height = gdk_pixbuf_get_height (src);
+	if (src_height < 0)
+		src_height = gdk_pixbuf_get_height (src);
 	
 	if (dest_x < 0)
 		dest_x = 0;
@@ -2185,17 +2362,17 @@ pixbuf_blend (GdkPixbuf *src,
 	if (dest_y < 0)
 		dest_y = 0;
 	
-	if (dest_x + width > dest_width) {
-		width = dest_width - dest_x;
+	if (dest_x + src_width > dest_width) {
+		src_width = dest_width - dest_x;
 	}
 	
-	if (dest_y + height > dest_height) {
-		height = dest_height - dest_y;
+	if (dest_y + src_height > dest_height) {
+		src_height = dest_height - dest_y;
 	}
 
 	gdk_pixbuf_composite (src, dest,
 			      dest_x, dest_y,
-			      width, height,
+			      src_width, src_height,
 			      offset_x, offset_y,
 			      1, 1, GDK_INTERP_NEAREST,
 			      alpha * 0xFF + 0.5);
@@ -2368,14 +2545,14 @@ handle_text (GMarkupParseContext *context,
 		fs->file = g_strdup (text);
 		slide->file1 = g_slist_prepend (slide->file1, fs);
 		if (slide->file1->next != NULL)
-			parser->changes_with_size = TRUE;                       
+			parser->has_multiple_sizes = TRUE;                       
 	}
 	else if (stack_is (parser, "size", "file", "static", "background", NULL) ||
 		 stack_is (parser, "size", "from", "transition", "background", NULL)) {
 		fs = slide->file1->data;
 		fs->file = g_strdup (text);
 		if (slide->file1->next != NULL)
-			parser->changes_with_size = TRUE; 
+			parser->has_multiple_sizes = TRUE; 
 	}
 	else if (stack_is (parser, "to", "transition", "background", NULL)) {
 		for (i = 0; text[i]; i++) {
@@ -2390,20 +2567,21 @@ handle_text (GMarkupParseContext *context,
 		fs->file = g_strdup (text);
 		slide->file2 = g_slist_prepend (slide->file2, fs);
 		if (slide->file2->next != NULL)
-			parser->changes_with_size = TRUE;                       
+			parser->has_multiple_sizes = TRUE;                       
 	}
 	else if (stack_is (parser, "size", "to", "transition", "background", NULL)) {
 		fs = slide->file2->data;
 		fs->file = g_strdup (text);
 		if (slide->file2->next != NULL)
-			parser->changes_with_size = TRUE;
+			parser->has_multiple_sizes = TRUE;
 	}
 }
 
-static void
+static SlideShow *
 slideshow_ref (SlideShow *show)
 {
 	show->ref_count++;
+	return show;
 }
 
 static void
@@ -2639,9 +2817,9 @@ get_thumb_annotations (GdkPixbuf *thumb,
 }
 
 static gboolean
-slideshow_changes_with_size (SlideShow *show)
+slideshow_has_multiple_sizes (SlideShow *show)
 {
-	return show->changes_with_size;
+	return show->has_multiple_sizes;
 }
 
 /*
@@ -2712,7 +2890,7 @@ gnome_bg_create_frame_thumbnail (GnomeBG			*bg,
 
 	result = gdk_pixbuf_new (GDK_COLORSPACE_RGB, FALSE, 8, dest_width, dest_height);
 
-	draw_color (bg, result);
+	draw_color (bg, result, screen);
 
 	thumb = create_img_thumbnail (bg, factory, screen, dest_width, dest_height, frame_num + skipped);
 
